#include <malloc.h>     // for _alloca
#include <excpt.h>      // for EXCEPTION_EXECUTE_HANDLER

#include "vars.h"

// Enable the strict_gs_check pragma

#ifdef GS_STRICT
  #pragma strict_gs_check(on)
#endif


// This function is used to prevent optimizations from removing the local
// variables from the test functions. It has to be defined in a different
// translation unit because otherwise the inter-procedural optimizations
// will figure out that it doesn't actually do anything.

void use(void* arg);


// Export the test functions and make the callee clear the stack

#define TEST __declspec(dllexport)


// http://msdn.microsoft.com/en-us/library/8dbf701c(VS.80).aspx
//
// The compiler injects checks in functions with local string buffers or, on
// x86, functions with exception handling. A string buffer is defined as an
// array whose element size is one or two bytes, and where the size of the
// whole array is at least five bytes, or, any buffer allocated with _alloca.

TEST void test_int() {
    int buf;

    use(&buf);
}

// The minimum size for an array is 5 chars or 3 shorts

TEST void test_array_of_4_char() {
    char buf[4];

    use(&buf);
}

TEST void test_array_of_5_char() {
    char buf[5];

    use(&buf);
}

TEST void test_array_of_2_short() {
    short buf[2];

    use(&buf);
}

TEST void test_array_of_3_short() {
    short buf[3];

    use(&buf);
}

// Typedefs should not confuse GS

typedef char tchar;

TEST void test_array_of_typedef_char() {
    tchar buf[5];

    use(&buf);
}

// Signedness should not confuse GS

TEST void test_array_of_signed_char() {
    signed char buf[5];

    use(&buf);
}

TEST void test_array_of_unsigned_char() {
    unsigned char buf[5];

    use(&buf);
}

TEST void test_array_of_signed_short() {
    signed short buf[5];

    use(&buf);
}

TEST void test_array_of_unsigned_short() {
    unsigned short buf[5];

    use(&buf);
}

// Arrays of integers and pointers are not protected by default

TEST void test_array_of_int() {
    int buf[5];

    use(&buf);
}

TEST void test_array_of_ptr() {
    char* buf[5];

    use(&buf);
}

// Arrays of structs and unions of size 1 or 2 bytes are protected

TEST void test_array_of_struct_1_char() {
    struct {
        char a;
    } buf[5];

    use(&buf);
}

TEST void test_array_of_struct_2_char() {
    struct {
        char a;
        char b;
    } buf[5];

    use(&buf);
}

TEST void test_array_of_struct_3_char() {
    struct {
        char a;
        char b;
        char c;
    } buf[5];

    use(&buf);
}

TEST void test_array_of_struct_1_short() {
    struct {
        short a;
    } buf[5];

    use(&buf);
}

TEST void test_array_of_struct_2_short() {
    struct {
        short a;
        short b;
    } buf[5];

    use(&buf);
}

TEST void test_array_of_struct_int() {
    struct {
        int a;
    } buf[5];

    use(&buf);
}

TEST void test_array_of_struct_char_int() {
    struct {
        char a;
        int b;
    } buf[5];

    use(&buf);
}

TEST void test_array_of_union_char_char() {
    union {
        char a;
        char b;
    } buf[5];

    use(&buf);
}

TEST void test_array_of_union_char_short() {
    union {
        char a;
        short b;
    } buf[5];

    use(&buf);
}

TEST void test_array_of_union_short_short() {
    union {
        short a;
        short b;
    } buf[5];

    use(&buf);
}

// Nested structs

TEST void test_array_of_struct_char_struct_char() {
    struct {
        struct {
            char a;
        } a;
        struct {
            char b;
        } b;
    } buf[5];

    use(&buf);
}

TEST void test_array_of_structs_char_struct_short() {
    struct {
        struct {
            char a;
        } a;
        struct {
            short b;
        } b;
    } buf[5];

    use(&buf);
}

// Structures are not protected by default

TEST void test_struct_1_char() {
    struct {
        char a;
    } buf;

    use(&buf);
}

TEST void test_struct_5_char() {
    struct {
        char a1;
        char a2;
        char a3;
        char a4;
        char a5;
    } buf;

    use(&buf);
}

// Structures and unions containing arrays are protected

TEST void test_struct_array_char() {
    struct {
        char a[5];
    } buf;

    use(&buf);
}

TEST void test_union_array_char_int() {
    union {
        char a[5];
        int b;
    } buf;

    use(&buf);
}

// Allocating memory with _alloca

TEST void test_alloca() {
    char* buf = _alloca(4);

    use(&buf);
}

// Structured exception handling

TEST int test_seh(int a, int b) {
    __try {
        return a / b;
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        return 0;
    }
}

TEST void test_seh_int(int a, int b) {
    int buf;

    __try {
        *(int*)&buf = a / b;
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        use(&buf);
    }
}

TEST void test_seh_array_4_char(int a, int b) {
    char buf[4];

    __try {
        *(int*)&buf = a / b;
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        use(&buf);
    }
}

TEST void test_seh_array_5_char(int a, int b) {
    char buf[5];

    __try {
        *(int*)&buf = a / b;
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        use(&buf);
    }
}

TEST void test_seh_array_250_char(int a, int b) {
    char buf[250];

    __try {
        *(int*)&buf = a / b;
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        use(&buf);
    }
}

TEST void test_seh_array_5000_char(int a, int b) {
    char buf[5000];

    __try {
        *(int*)&buf = a / b;
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        use(&buf);
    }
}

// Argument tests

TEST void test_arg_int(int arg) {
    use(&arg);
}

TEST void test_arg_ptr(char* arg) {
    use(&arg);
}

TEST void test_arg_array_char(char arg[5]) {
    use(&arg);
}

typedef struct {
    int a;
} struct_int;

TEST void test_arg_struct_int(struct_int arg) {
    use(&arg);
}

typedef struct {
    char a[4];
} struct_array_4_char;

TEST void test_arg_struct_array_4_char(struct_array_4_char arg) {
    use(&arg);
}

typedef struct {
    char a[5];
} struct_array_5_char;

TEST void test_arg_struct_array_5_char(struct_array_5_char arg) {
    use(&arg);
}


//
// Variable reordering
//

TEST vars_t* test_reorder_locals() {
    int   a;
    void* b;
    char  c[4];
    char  d[5];
    int   e[5];
    void* f[10];
    struct { int a; } g;
    struct { int a; int b; } h;
    struct { char a[10]; } i;
    struct { int a; int b; int c; int d; int e; int f; } j;

    int   k;
    void* l;
    char  m[4];
    char  n[5];
    int   o[5];
    void* p[10];
    struct { int a; } q;
    struct { int a; int b; } r;
    struct { char a[10]; } s;
    struct { int a; int b; int c; int d; int e; int f; } t;


    return vars(20,
        &a,  "int   a",
        &b,  "void* b",
        &c,  "char  c[4]",
        &d,  "char  d[5]",
        &e,  "int   e[5]",
        &f,  "void* f[10]",
        &g,  "{ 1 int } g",
        &h,  "{ 2 ints } h",
        &i,  "{ char[10]; } i",
        &j,  "{ 6 ints } j",
        &k,  "int   k",
        &l,  "void* l",
        &m,  "char  m[4]",
        &n,  "char  n[5]",
        &o,  "int   o[5]",
        &p,  "void* p[10]",
        &q,  "{ 1 int } q",
        &r,  "{ 2 ints } r",
        &s,  "{ char[10]; } s",
        &t,  "{ 6 ints } t"
    );
}

TEST vars_t* test_reorder_locals_order() {
    char a[10];
    struct { char a[30]; } b;
    char c[50];
    struct { char a[70]; } d;
    char e[80];
    struct { char a[60]; } f;
    char g[40];
    struct { char a[20]; } h;

    char i;
    int j;
    struct { char a[2]; } k;
    void* l[2];
    char m[3];
    int n[10];
    struct { char a; char b; } o;
    
    return vars(15,
        &a, "char[10] a",
        &b, "{ char[30] } b",
        &c, "char[50] c",
        &d, "{ char[70] } d",
        &e, "char[80] e",
        &f, "{ char[60] } f",
        &g, "char[40] g",
        &h, "{ char[20] } h",
        &i, "char i",
        &j, "int j",
        &k, "{ char[2] } k",
        &l, "void* l[2]",
        &m, "char m[3]",
        &n, "int n[10]",
        &o, "{ char; char; } o"
    );
}

// Disable frame pointer oprimizations

#pragma optimize("y", off)

__declspec(naked) __declspec(noinline) void* get_ebp()
{
    __asm {
        mov eax, ebp
        ret
    }
}

TEST vars_t* test_reorder_args(int   a,
                               char* b,
                               struct { int a; char* b; } c,
                               struct { char a[2]; } d,
                               struct { char a[5]; } e,
                               struct { char* a[5]; } f
                              ) {
    int   local1;
    char  local2[5];

    return vars(9,
        &a, "int   a",
        &b, "char* b",
        &c, "{ int; char*; } c",
        &d, "{ char[2]; } d",
        &e, "{ char[5]; } e",
        &f, "{ char*[5]; } f",
        get_ebp(), "-----------",
        &local1,   "int  local1",
        &local2,   "char local2[5]"
    );
}

TEST vars_t* test_reorder_strict_bug(struct { char a[5]; } a)
{
    return vars(2,
        get_ebp(), "-----------",
        &a, "{ char[5] } a"
    );
}
