//
// seh-test.c
//
// The seh-test program displays the current process flags and tries to execute
// an exception handler in the text segment, data segment, heap memory or the
// stack. It is compiled with different permutations of the /SafeSEH and
// /NXcompat flags and can be used to test the SEH handler validation in
// programs with different combinations of DEP and SafeSEH.

#include <stdio.h>

#include <windows.h>
#include <excpt.h>

#define _WIN32_WINNT 0x0500
#include <winternl.h>


//
// Returns the address of an exported function in a DLL
//

FARPROC resolve_func(char* dll, char* func)
{
    HMODULE module;
    FARPROC addr;
    
    module = LoadLibrary(dll);
    if (module == NULL) {
        printf("Failed to load DLL %s\n", dll);
        exit(1);
    }

    addr = GetProcAddress(module, func);
    if (addr == NULL) {
        printf("Failed to resolve function %s in DLL %s\n", func, dll);
        exit(1);
    }

    return addr;
}


const DWORD ProcessExecuteFlags = 0x22;

typedef struct {
    unsigned int ExecuteDisable : 1;
    unsigned int ExecuteEnable : 1;
    unsigned int DisableThunkEmulation : 1;
    unsigned int Permanent : 1;
    unsigned int ExecuteDispatchEnable : 1;
    unsigned int ImageDispatchEnable : 1;
    unsigned int DisableExceptionChainValidation : 1;
    unsigned int Spare : 25;
} KEXECUTE_OPTIONS;


//
// Returns the process execution flags
//

KEXECUTE_OPTIONS get_process_flags()
{
    FARPROC NtQueryInformationProcess;
    NTSTATUS status;
    KEXECUTE_OPTIONS flags;
    DWORD len;

    NtQueryInformationProcess = resolve_func("ntdll.dll",
        "NtQueryInformationProcess");

    status = NtQueryInformationProcess(GetCurrentProcess(),
                                       ProcessExecuteFlags,
                                       &flags,
                                       sizeof(flags),
                                       &len);

    if (status < 0) {
        printf("NtQueryInformationProcess returned error code 0x%x\n", status);
        exit(1);
    }

    if (len != 4) {
        printf("NtQueryInformationProcess returned unexpected length %d\n",
               len);
        exit(1);
    }

    return flags;
}


//
// Display the procoess execution flags
//

void dump_process_flags()
{
    KEXECUTE_OPTIONS flags = get_process_flags();

    printf("Process execution flags:\n"
        "    ExecuteDisable        : %d\n"
        "    ExecuteEnable         : %d\n"
        "    DisableThunkEmulation : %d\n"
        "    Permanent             : %d\n"
        "    ExecuteDispatchEnable : %d\n"
        "    ImageDispatchEnable   : %d\n"
        "    DisableExceptionChainValidation : %d\n\n",
        flags.ExecuteDisable,
        flags.ExecuteEnable,
        flags.DisableThunkEmulation,
        flags.Permanent,
        flags.ExecuteDispatchEnable,
        flags.ImageDispatchEnable,
        flags.DisableExceptionChainValidation);
}


//
// Test an SEH handler
//

void seh_test(FARPROC handler)
{
    EXCEPTION_POINTERS *info;

    struct EXCEPTION_REGISTRATION {
        DWORD prev;
        DWORD handler;
    } record;

    DWORD crash_addr;   // address of the instruction that causes the exception

    __try {

        record.handler = (DWORD)handler;

        __asm {
            // save the old handler

            mov eax, fs:[0]
            mov record.prev, eax

            // get the address of the instruction that will cause an exception

            lea eax, crash
            mov crash_addr, eax

            // install the new handler

            lea eax, record
            mov fs:[0], eax

            // cause an access violation exception

            xor eax, eax
crash:
            mov [eax], eax
        }

        printf("    handler executed\n");
    }
    __except(info = GetExceptionInformation(), EXCEPTION_EXECUTE_HANDLER) {

        __asm {
            int 3;
        }

        if (info->ExceptionRecord->ExceptionCode == 0xc0000005 &&
            info->ContextRecord->Eip == crash_addr)
        {
            printf("    handler not executed, but the search for SEH handlers continued\n");
        }
        else {
            printf("    handler not executed, but we got an unexpected second exception\n");

            printf("    exception code: 0x%x\n",
                info->ExceptionRecord->ExceptionCode);

            printf("    eip: 0x%x\n", info->ContextRecord->Eip);
        }

        __asm {
            // restore the old handler

            mov eax, record.prev
            mov fs:[0], eax
        }
    }
}


//
// Exception handler in the text segment, but not in the SafeSEH table
//

EXCEPTION_DISPOSITION __cdecl handler_text(
    struct _EXCEPTION_RECORD *ExceptionRecord,
    void* EstablisherFrame,
    struct _CONTEXT *ContextRecord,
    void* DispatcherContext)
{
    // Adjust eip to skip over the faulting mov [eax], eax instruction

    ContextRecord->Eip += 2;

    return ExceptionContinueExecution;
}


//
// Exception handler in the data segment
//

char handler_data[] =

    // ContextRecord->Eip += 2;
    "\x8B\x44\x24\x0C"              // mov     eax, [esp+0Ch]
    "\x83\x80\xB8\x00\x00\x00\x02"  // add     dword ptr [eax+0B8h], 2

    // return ExceptionContinueExecution;
    "\x33\xC0"                      // xor eax, eax
    "\xC3";                         // retn

//
// Program usage
//

void usage() {
    printf("Syntax: seh-test <test>\n\n");
    printf("Tests:\n");
    printf("  text    handler in a text segment, but not in the SafeSEH table\n");
    printf("  data    handler in a data segment\n");
    printf("  heap    handler on the heap\n");
    printf("  stack   handler on the stack\n");
}


//
// Main function
//

int main(int argc, char* argv[])
{
    char handler_stack[sizeof(handler_data)];
    char* handler_heap;

    if (argc != 2) {
        usage();
        return 1;
    }

    if (strcmp(argv[1], "text") == 0) {

        dump_process_flags();

        // Handler in a text segment but not in the SafeSEH table

        printf("Handler in a text segment:\n");

        seh_test((FARPROC)handler_text);
    }
    else if (strcmp(argv[1], "data") == 0) {

        dump_process_flags();

        // Handler in a data segment

        printf("Handler in a data segment:\n");

        seh_test((FARPROC)handler_data);
    }
    else if (strcmp(argv[1], "heap") == 0) {

        dump_process_flags();

        // Handler on the heap

        handler_heap = malloc(sizeof(handler_data));
        if (handler_heap == NULL) {
            printf("Error allocating %d bytes for the handler on the heap\n",
                sizeof(handler_data));
            exit(1);
        }

        memcpy(handler_heap, handler_data, sizeof(handler_data));

        printf("Handler on the heap:\n");

        seh_test((FARPROC)handler_heap);
    }
    else if (strcmp(argv[1], "stack") == 0) {

        dump_process_flags();

        // Handler on the stack

        memcpy(handler_stack, handler_data, sizeof(handler_data));

        printf("Handler on the stack:\n");

        seh_test((FARPROC)handler_stack);
    }
    else {
        usage();
        return 1;
    }

    return 0;
}
